# encoding: utf-8

from datetime import date

from auf.django.references import models as ref
from django.db import models
from django.db.models import Sum
from django.db.models.query import QuerySet


class Element(models.Model):
    code = models.CharField(max_length=72, primary_key=True)
    timestamp = models.BigIntegerField()
    nom = models.CharField(max_length=36)

    class Meta:
        abstract = True
        ordering = ['code']

    def __unicode__(self):
        return self.code


class PCG(Element):

    class Meta:
        db_table = 'coda_pcg'


class TiersOperation(Element):
    etablissement = models.OneToOneField(
        ref.Etablissement, related_name='coda_tiers_operation',
        blank=True, null=True
    )

    class Meta:
        db_table = 'coda_tiersoperation'

    def solde_cotisations_par_annee(self):
        """
        Retourne une liste de paires (annee, solde) contenant le solde des
        cotisations à payer par année pour cet établissement.
        """
        cotisations_impayees = self.lignes.cotisations().exclude(
            etat_paiement='P'
        )
        factures_par_annee = list(
            cotisations_impayees
            .filter(montant_eur__gt=0)
            .values('ecriture__annee')
            .annotate(Sum('montant_eur'))
            .order_by('-ecriture__annee')
        )
        paiements = (
            cotisations_impayees
            .filter(montant_eur__lt=0)
            .aggregate(Sum('montant_eur'))
        )['montant_eur__sum'] or 0

        while paiements < 0 and len(factures_par_annee) > 0:
            paiements += factures_par_annee[-1]['montant_eur__sum']
            if paiements <= 0:
                factures_par_annee.pop()
        if paiements > 0 and len(factures_par_annee) > 0:
            # Ajuster le solde à payer sur la dernière facture
            factures_par_annee[-1]['montant_eur__sum'] = paiements
        return [(x['ecriture__annee'], x['montant_eur__sum'])
                for x in factures_par_annee]

    def delinquant(self):
        """
        Teste si l'établissement lié à ce tiers est délinquant ou pas.

        Un établissement délinquant a 3 années ou plus de cotisations
        impayées.
        """
        today = date.today()
        annee_max = today.year
        if today.month < 6:
            # Ne pas considérer l'année en cours avant le mois de juin
            annee_max -= 1
        return len([x for x in self.solde_cotisations_par_annee()
                    if x[0] <= annee_max]) >= 3


class ProjetPoste(Element):

    class Meta:
        db_table = 'coda_projetposte'


class Ecriture(models.Model):
    id = models.CharField(max_length=25, primary_key=True)
    code = models.CharField(max_length=12)
    numero = models.CharField(max_length=12)
    timestamp = models.BigIntegerField()
    periode = models.IntegerField()
    annee = models.IntegerField()
    date = models.DateField()

    class Meta:
        db_table = 'coda_ecriture'
        ordering = ['id']


class LigneQuerySet(QuerySet):

    def cotisations(self):
        """
        Filtre qui ne garde que les lignes qui représentent des factures ou
        des paiements de cotisations d'établissements membres.
        """
        return self \
                .exclude(tiers_operation__etablissement=None) \
                .exclude(montant_eur=0) \
                .exclude(ecriture__code='G-CLOT-BILAN') \
                .exclude(ecriture__code='C-ENC-INSCR') \
                .exclude(ecriture__code='G-ODD',
                         ecriture__numero__in=(19912, 29582))


class LigneManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self):
        return LigneQuerySet(self.model)

    def cotisations(self):
        return self.get_query_set().cotisations()


class Ligne(models.Model):
    DEBIT_CREDIT_CHOICES = (
        ('D', 'Débit'),
        ('C', 'Crédit'),
    )
    ETAT_PAIEMENT_CHOICES = (
        ('D', 'Disponible'),
        ('A', 'Attente'),
        ('X', 'Fermé (non-lettrable)'),
        ('P', 'Payé'),
        ('O', 'Proposé'),
        ('F', 'Effet en cours'),
        ('Y', 'Effet payé'),
        ('J', 'Effet en attente'),
    )

    id = models.CharField(max_length=36, primary_key=True)
    ecriture = models.ForeignKey(Ecriture, related_name='lignes')
    numero = models.IntegerField()
    timestamp = models.BigIntegerField()
    pcg = models.ForeignKey(PCG, related_name='lignes')
    tiers_operation = models.ForeignKey(
        TiersOperation, related_name='lignes', blank=True, null=True
    )
    projet_poste = models.ForeignKey(
        ProjetPoste, related_name='lignes', blank=True, null=True
    )
    description = models.CharField(max_length=36)
    montant_eur = models.DecimalField(max_digits=17, decimal_places=2)
    debit_credit = models.CharField(max_length=1, choices=DEBIT_CREDIT_CHOICES)
    numero_facture = models.CharField(max_length=32)
    numero_avoir = models.CharField(max_length=32)
    implantation = models.CharField(max_length=32)
    salarie = models.CharField(max_length=32)
    numero_cheque = models.CharField(max_length=32)
    etat_paiement = models.CharField(
        max_length=1, choices=ETAT_PAIEMENT_CHOICES
    )

    # Managers
    objects = LigneManager()

    class Meta:
        db_table = 'coda_ligne'
        ordering = ['id']
